import { ConsoleLogger, Injectable } from '@nestjs/common';
import * as fs from 'fs';
import * as path from 'path';
import * as crypto from 'crypto';
import * as child_process from 'child_process';
import * as Mustache from 'mustache';
import tar from 'tar';
import { App } from '../entities/App.entity';
import PQueue from 'p-queue';

// eslint-disable-next-line @typescript-eslint/no-empty-function
function nothing() {}

@Injectable()
export class PackagerService extends ConsoleLogger {
  workingPath: string;
  releasePath: string;
  downloadBaseUrl: string;
  queueIdMap = new Map<string, PQueue>();

  getQueue(id: string) {
    if (!this.queueIdMap.has(id)) {
      this.queueIdMap.set(id, new PQueue({ concurrency: 1 }));
    }
    return this.queueIdMap.get(id);
  }

  constructor() {
    super('packager');
  }

  private async readdirRecursive(_path: string): Promise<string[]> {
    const files = await fs.promises.readdir(_path, { encoding: 'utf-8' });
    let result = files;
    for (const file of files) {
      const child = path.join(_path, file);
      const stat = await fs.promises.stat(child);
      if (stat.isDirectory()) {
        result = result.concat(
          (await this.readdirRecursive(child)).map((_file) =>
            path.join(file, _file),
          ),
        );
      }
    }
    return result;
  }

  private caculateSHA256(file: string): Promise<string> {
    return new Promise((resolve, reject) => {
      const input = fs.createReadStream(file);
      const hash = crypto.createHash('sha256');
      hash.on('error', (error: Error) => {
        reject(error);
      });
      input.on('error', (error: Error) => {
        reject(error);
      });
      hash.on('readable', () => {
        const data = hash.read();
        if (data) {
          resolve((<Buffer>data).toString('hex'));
        }
      });
      input.pipe(hash);
    });
  }

  private spawnAsync(
    command: string,
    args: string[],
    options: child_process.SpawnOptions,
  ) {
    return new Promise<void>((resolve, reject) => {
      const child = child_process.spawn(command, args, options);
      child.on('exit', (code) => {
        if (code == 0) {
          resolve();
        } else {
          reject(code);
        }
      });
      child.on('error', (error) => {
        reject(error);
      });
    });
  }

  private archive(
    archive: string,
    files: string[],
    directory: string,
  ): Promise<void> {
    return tar.c({ gzip: true, C: directory, file: archive }, files);
  }

  private async createDirectoryIfNotExists(directory: string) {
    try {
      await fs.promises.access(directory);
    } catch (e) {
      await fs.promises.mkdir(directory, { recursive: true });
    }
  }

  private async unarchiveStream(stream: fs.ReadStream, directory: string) {
    await this.createDirectoryIfNotExists(directory);
    const unarchiveProcessStream = stream.pipe(tar.x({ C: directory }));
    return new Promise<void>((resolve, reject) => {
      unarchiveProcessStream.on('finish', resolve);
      unarchiveProcessStream.on('error', reject);
    });
  }

  private async packageProcess(app: App, stream: fs.ReadStream) {
    const unarchiveDestination = path.join(
      this.workingPath,
      app.packageFullPath,
    );
    const packageWorkingRoot = path.join(this.workingPath, app.id);
  }

  async package(app: App, stream: fs.ReadStream) {
    const queue = this.getQueue();
  }
}
